#
# Alternatives.pm -- handle the Debian alternatives system
# By Jaldhar H. Vyas <jaldhar@braincells.com>
# Copyright (c) 2003, Consolidated Braincells Inc. 
# See COPYING in the source distribution for license details.
#

# This is a subclass of CGI::Application
package Alternatives;
use base 'CGI::Application';

use strict;

# unfortunately webmin's web-lib.pl uses a lot of global functions so because 
# we use strict we have to predeclare them here.
use vars qw(%config %gconfig $module_name $module_config_directory $tb $cb 
  $scriptname $remote_user $base_remote_user $current_theme $root_directory 
  $module_root_directory %module_info);


# This function is the first one called when the script is run.  Ideally this 
# and cgiapp_postrun would be in a class which contained all the infrastructure
# that webmin modules using CGI::Application share in common.
#  The inheritance tree would look like this:
#
#                                                 ,---> Alternatives.pm
#  CGI::Application --> CGI::Application::Webmin -+---> Foo.pm
#                       (or some similiar name)   `---> Bar.pm
#                                                       (etc.)
sub cgiapp_init
{
  my ($self) = @_;

  # Import the webmin helper functions.
  do '../web-lib.pl';

  # initalize webmin configuration.
  init_config();

  # Convert all the webmin global variables into class parameters to make
  # it more OOP.
  map { $self->param($_, $config{$_}) } keys %config;
  map { $self->param($_, $config{$_}) } keys %gconfig;
  map { $self->param($_, $config{$_}) } keys %module_info;
  $self->param('module_name', $module_name);
  $self->param('module_config_directory', $module_config_directory);
  $self->param('tb', $tb);
  $self->param('cb', $cb);
  $self->param('scriptname', $scriptname);
  $self->param('remote_user', $remote_user);
  $self->param('base_remote_user', $base_remote_user);
  $self->param('current_theme', $current_theme);
  $self->param('root_directory', $root_directory);
  $self->param('module_root_directory', $module_root_directory);
}


#
# This function is called after each run mode (See CGI::Application docs.)
# Here we wrap the output of each template with the webmin header and footer
#
sub cgiapp_postrun
{
  # $output is a reference to the results of a run mode.
  my ($self, $output) = @_;

  # We want to send out the input unbuffered.
  local $| = 1;

  # We don't want CGI::Application to send the http headers.  The function
  # below will do that.
  $self->header_type('none');

  # print the webmin header.  
  header($self->param('module_name'), '');

  print $$output;

  # print the webmin footer
  footer();

  # If we don't erase the contents of $output, the body of the page will be
  # displayed twice
  $$output = '';
}

#
# Here we set up the Alternatives class.  Think of it as the constructor.
#
sub setup
{
  my ($self) = @_;

  # The initial run mode or if no run mode is specified.
  $self->start_mode('display_all');

  #  Our run modes and the functions they map to.
  $self->run_modes(
    'display_all' => 'display_all',
    'list' => 'list',
    # AUTOLOAD is called if something other than the above run modes is asked
    # for.  In this case we just fall back to display_all. 
    AUTOLOAD => 'display_all',
  );

  # /usr/sbin/update-alternatives does this.  I don't know why but we'll do it 
  # too. :-)
  my $enoent = readpipe($self->param('dpkglibdir') . '/enoent') or
    terror('enoent', $self->param('dpkglibdir'), $!);
  $self->param('ENOENT', $enoent);

}

#
#  This run mode displays all the alternatives and their status.
#
sub display_all
{
  my ($self) = @_;

  my @alternatives;

  opendir (DIR, $self->param('admindir')) 
    or terror('fail_diropen', $self->param('admindir'), $!);

  while (defined (my $alternative = readdir DIR))
  {
    next if $alternative =~ /\.{1,2}/;

    push @alternatives, $self->parse_alternative(\$alternative);
  }
  @alternatives = sort { $a->{name} cmp $b->{name} } @alternatives;
  closedir DIR;

  # See the HTML::Template man page to understand all this.
  my $template = $self->load_tmpl('display', die_on_bad_params => 0);
  $template->param(alternatives => \@alternatives, tb => $self->param('tb'),
    cb => $self->param('cb'), scriptname => $self->param('scriptname'),);
  return $template->output;
}

#
#  This run mode displays the information for one alternative group.
#
sub list
{
  my ($self) = @_;
  my $alternative = $self->query->param('name');

  my $file = $self->parse_alternative(\$alternative);
  my $template = $self->load_tmpl('list', die_on_bad_params => 0);
  $template->param(name => $file->{name}, links => $file->{links},
    scriptname => $self->param('scriptname'), tb => $self->param('tb'),
    cb => $self->param('cb'),);

  return $template->output;
}

#
#  This is a helper function that parses an alternative.
#  The code mostly comes from /usr/sbin/update-alternatives.

sub parse_alternative
{
  my ($self, $name) = @_;
 
  # getline function implemented as a closure for efficiency.
  my $getline = sub
  {
    my ($self, $item) = @_;
    $!=0; 
    $_= <FILE>;
    length($_) or
      terror('getline1', $self->param('admindir'), $$name, $item, $!);
    s/\n$// or terror('getline2', $item);
    return $_;
  };

  my ($manual, $link);

  if (open(FILE, '<', $self->param('admindir') . "/$$name"))
  {
    $manual = $getline->('manflag');
    $manual = text('badformat') unless $manual eq 'auto' || $manual eq 'manual';
    $link = $getline->('link');
   
    my (@slavenames, %slavenum, %slavelinkcount, @slavelinks);

    while ((my $sname = $getline->('sname')) ne '')
    {
      push(@slavenames,$sname);
      terror('duplicate_slave_name', $sname) if defined($slavenum{$sname});
      $slavenum{$sname} = $#slavenames;
      my $slink = $getline->('slink');
      terror('samelink', $link) if $slink eq $link;
      terror('duplicate_slave_link', $slink) if $slavelinkcount{$slink}++;
      push(@slavelinks,$slink);
    }

    my (@versions, %versionnum, %priorities, @priorities, %slavepath);
    while ((my $version = $getline->('version')) ne '')
    {
      terror('duplicate_path', $version) if defined($versionnum{$version});
      if ( -r $version )
      {
        my $i;
        push(@versions,$version);
        $versionnum{$version} = $i= $#versions;
        my $priority= $getline->('priority');
        terror('priority',$version,$priority) unless $priority =~ /^[-+]?\d+$/;
        $priorities[$i] = $priority;
        for (my $j=0; $j <= $#slavenames; $j++)
        {
          $slavepath{$i,$j} = $getline->('spath');
        }
      }
      else
      {
        # File not found - remove
        $getline->('priority');
        for (my $j=0; $j <= $#slavenames; $j++)
        {
          $getline->('spath');
        }
      }
    }

    close FILE;

    my @links;

    foreach (keys %slavenum)
    {
      $links[$slavenum{$_}]->{name} = $_;
      $links[$slavenum{$_}]->{path} = $slavelinks[$slavenum{$_}];
    }

    unshift (@links, { name => $$name, path => $link });

    foreach my $version (@versions)
    {
      
      push @{$links[0]->{versions}}, {version => $version, 
        priority => @priorities[$versionnum{$version}] };
     
      for (my $j=0; $j <= $#slavenames; $j++)
      {
        push @{$links[$j + 1]->{versions}}, {
          version => $slavepath{$versionnum{$version},$j}, 
          priority => undef };
      }
    }

    return {name => $$name, manflag => $manual, links => \@links, };
  }
  elsif ($! != $self->param('ENOENT'))
  {
    terror('fail_open', $self->param('admindir'), $$name, $!);
  }
}


1;

__END__

